/*******************************************************************************
**		This DLL includes functions that allow Windows 95 applications to       **
**		perform I/O operations both to physical memory and I/O ports.          	**
**                                                                            **
**    B&B Technologies Ltd. Co.                                               **
**    6610 Gulton Court NE			             Created by: Juan Ortiz-Moyet    **
**		Albuquerque, NM 87109-1477              Last modification:              **
**		(505) 345-9499                           Added support for PCI devices	**
**    Fax: (505)345-9699                       											**
**		e-mail: bbtech@roadrunner.com				 Made on: July 17, 1998				**
**															 By: Juan Ortiz-Moyet            **
*******************************************************************************/

#include "sat95.h"

/* Global variables	*/
	/* Queue of active session resources.	*/
static struct Table Session[MAX_SESSION_NUMBER];

	/* Array of Driver initialization flags.	*/
static BOOL FirstSession[MAX_SESSION_NUMBER];

	/* Interrupt session flag.	*/
static DWORD ID;

/*******************************************************************************
**	   DllEntryPoint():	Function to initialize and terminate DLL instance		**
**																										**
**		Arguments:  check Win32 function reference for details						**
**																										**
**		Return value:	Success = TRUE															**
**							Failure = FALSE							 							**
**										 (specific error code - GetLastError() )			**
*******************************************************************************/
BOOL WINAPI DllEntryPoint(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
 register DWORD 	i;

 switch (fdwReason)
   {
    case DLL_PROCESS_ATTACH:
      for ( i=0; i<MAX_SESSION_NUMBER; i++)  /* When the DLL is first loaded   */
        {                                    /* by any process, it must        */
         Session[i].Handle = NULL;          	/* clear the contents of the      */
         Session[i].Device = INVALID_DEVICE_NUMBER;
         Session[i].PhysMem1 = NULL;         /* Session Table for the code to  */
         Session[i].PhysMem2 = NULL;         /* execute without errors.			 */
			Session[i].Occurrence = -1;
         Session[i].WndHandle = NULL;
         Session[i].IsrEvent = NULL;
         Session[i].TerminateIsr = FALSE;
      	FirstSession[i] = TRUE;   /* Signal initialization of driver session. */
        }
      	/* Initialize interrupt session Id.	*/
      ID = 0;
      break;

    case DLL_PROCESS_DETACH:	/* Unload ALL VxDs.  */
    	{
 		 HANDLE  	hVxDLdr;    					/* Handle to VxD Loader */
		 IOCTL_PKT 	DL_IOCTL_Pkt;
		 DWORD		dwBytesReturned;
       USHORT		usErrorCode;
       register i;

			/* Get file handle to VXDLDR VxD */
		 hVxDLdr = CreateFile("\\\\.\\VXDLDR", 0,0,0,0,0,0);
		 if ( hVxDLdr == INVALID_HANDLE_VALUE )
	    	return( TRUE );

       for ( i = 0; i < MAX_NUMBER_OF_VXD; i++)
       	{
        		/* Create the Module name (this is case sensitive). */
	    	 sprintf((char *)(DL_IOCTL_Pkt.Res_DL_Pkt.W32IO_ModuleName), "WRTDEV%d", i);
		 	 DL_IOCTL_Pkt.Res_DL_Pkt.W32IO_DeviceID = (USHORT)-1;

		      /* Call VxD Loader to Unload our Vxd. */
		    DeviceIoControl(hVxDLdr, VXDLDR_APIFUNC_UNLOADDEVICE, (LPVOID)&DL_IOCTL_Pkt,
     						sizeof(IOCTL_PKT), (LPVOID)&usErrorCode, sizeof(USHORT),
            			&dwBytesReturned, NULL);
         }

			/* Done with the Loader. */
		 CloseHandle(hVxDLdr);

		 break;
		} /* DLL_PROCESS_DETACH case */

    case DLL_THREAD_ATTACH:
    case DLL_THREAD_DETACH:
    default: break;
   } /* fdwReason switch */

 return(TRUE);
}


/*******************************************************************************
**	   SessionControl():	Function to initialize and terminate I/O session		**
**																										**
**		Arguments:  pdwSessionID - number that describes I/O session; if eTask	**
**						is Start_Session, it is equal to the device to be accessed.	**
**						eTask - type of operation to perform								**
**																										**
**		Return value:	Success = 0																**
**							Failure = standard Windows error code 							**
**										 (defined in winerror.h)								**
*******************************************************************************/
DWORD SessionControl(uInt32 *pdwSessionID, enum SCtrlFunc eTask)
{
 DWORD dwErrorCode = ERROR_SUCCESS;
 DWORD dwDeviceNumber;
 char szDeviceName[MAX_WINRT_NAME];
 register int i;

 switch (eTask)
  {
   case (Start_Session):

     if (FirstSession[*pdwSessionID])
     	 {
// If device is PCI we must start the driver manually.
        DriverControl(Start_Driver,*pdwSessionID);
        FirstSession[*pdwSessionID] = FALSE;
//
 MessageBox(NULL, "First session for VxD.", NULL, MB_ICONEXCLAMATION | MB_OK);
//
       }

     dwDeviceNumber = *pdwSessionID;

 			/*	Search for an available index in Session table.	*/
     i = 1;					 /* Do not touch index 0; that one is always invalid.	*/
     while ( (Session[i].Device != INVALID_DEVICE_NUMBER) &&
 		  ( i < MAX_SESSION_NUMBER) ) i++;

     if ( Session[i].Device != INVALID_DEVICE_NUMBER )
     	 {
        *pdwSessionID = 0;       /* Invalidate SessionID.	*/
        return( 1 );      			/* Set error: no more devices are available.	*/
       }

         /* Save available index as new SessionID.	*/
     *pdwSessionID = i;

			/*	Generate DeviceName string with given device number.	*/
         /* Win NT device is WinRTdev*; Win95 device is WRTdev*.	*/
     sprintf(szDeviceName, "\\\\.\\WRTdev%d", dwDeviceNumber);

//
 MessageBox(NULL, (LPCTSTR)(szDeviceName), NULL, MB_ICONEXCLAMATION | MB_OK);
//

			/*	Obtain handle to new session.	*/
     Session[*pdwSessionID].Handle = CreateFile( (LPCTSTR)(szDeviceName), 0,
     					FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 0, NULL);

     if ( Session[*pdwSessionID].Handle == INVALID_HANDLE_VALUE )
     	{                                        /* Check for invalid handle */
    	 Session[*pdwSessionID].Handle = NULL;
       Session[*pdwSessionID].Device = INVALID_DEVICE_NUMBER;
       *pdwSessionID = 0;                      /* Invalidate SessionID		*/
//
 MessageBox(NULL, "Failed to open device handle.", NULL, MB_ICONEXCLAMATION | MB_OK);
//
    	 return( GetLastError() );
   	}

     		/*	Update the Session table to reflect the new device.	*/
     Session[*pdwSessionID].Device = dwDeviceNumber;
//
 MessageBox(NULL, "Succesfully opened device handle.", NULL, MB_ICONEXCLAMATION | MB_OK);
//
     break;

   case (End_Session):

     /* Close handle to I/O device.	 */
     if ( !CloseHandle(Session[*pdwSessionID].Handle) )
       dwErrorCode = GetLastError();  				/* Catch error on API call. */

     Session[*pdwSessionID].Handle = NULL;
     Session[*pdwSessionID].Device = INVALID_DEVICE_NUMBER;

		/* If session started an IsrFunction thread, terminate the thread.	*/
     if (Session[*pdwSessionID].IsrEvent != NULL )
     	{
       Session[*pdwSessionID].TerminateIsr = TRUE;
       SetEvent(Session[*pdwSessionID].IsrEvent);
      }

      /* First element of Session table always contains invalid data.	*/
     *pdwSessionID = 0;			/* Invalidate SessionID		*/
     break;

   default: break;
  }

 return( dwErrorCode );
}


/*******************************************************************************
**	   ByteIo():	Function to execute byte memory I/O operations					**
**																										**
**		Arguments:  dwSessionID - number that describes I/O session					**
**						iTask - specifies operation to perform								**
**									  ( TRUE => Read, FALSE => Write )						**
**						ulOffset - physical address to manage								**
**						ulNumBytes - number of BYTES to manage								**
**						pbData - address to hold requested byte(s)		 				**
**						iResGroup - especifies resource group for transfer				**
**									  ( TRUE => Secondary, FALSE => Primary )				**
**																										**
**		Return value:	ERROR_SUCCESS (See Winerror.h)									**
*******************************************************************************/
DWORD ByteIo(DWORD dwSessionID, BOOL iTask, ULONG ulOffset,
					ULONG ulNumBytes, PBYTE pbData, BOOL iResGroup)
{
 ULONG  i;
 PBYTE  pbAddress;

 if ( !iResGroup )
 	pbAddress = (PBYTE)(Session[dwSessionID].PhysMem1);
 else
   pbAddress = (PBYTE)(Session[dwSessionID].PhysMem2);

 if (pbAddress == NULL)              /* Check for invalid pointer before	*/
 	return ( ERROR_INVALID_HANDLE );  /* attempting any access.					*/

 pbAddress += ulOffset;     			 /* Start access at requested offset. 	*/

 if ( iTask )
  {
   	/* Read byte(s) into given location. 	*/
   for ( i = 0; i < ulNumBytes; i++ )
		*( pbData + i ) = *( pbAddress + i );
  }
 else
  {
   	/* Write byte(s) to specified location.	*/
   for ( i = 0; i < ulNumBytes; i++ )
		*( pbAddress + i ) = *( pbData + i );
  }

 return( ERROR_SUCCESS );
}


/*******************************************************************************
**	   WordIo():	Function to execute 16-bit memory I/O operations				**
**																										**
**		Arguments:  dwSessionID - number that describes I/O session					**
**						iTask - specifies operation to perform								**
**									  ( TRUE => Read, FALSE => Write )						**
**						ulOffset - physical address to manage								**
**						ulNumBytes - number of BYTES to manage								**
**						pwData - address to hold requested word(s)		 				**
**						iResGroup - especifies resource group for transfer				**
**									  ( TRUE => Secondary, FALSE => Primary )				**
**																										**
**		Return value:	ERROR_SUCCESS (See Winerror.h)									**
*******************************************************************************/
DWORD WordIo(DWORD dwSessionID, BOOL iTask, ULONG ulOffset,
					ULONG ulNumBytes, PUSHORT pwData, BOOL iResGroup)
{
 ULONG	 i;
 PUSHORT	 pwAddress;

 if ( !iResGroup )
 	pwAddress = (PUSHORT)(Session[dwSessionID].PhysMem1);
 else
   pwAddress = (PUSHORT)(Session[dwSessionID].PhysMem2);

 if (pwAddress == NULL)              /* Check for invalid pointer before	*/
 	return ( ERROR_INVALID_HANDLE );  /* attempting any access.					*/

 pwAddress += ulOffset;     			 /* Start access at requested offset. 	*/

 if ( iTask )
  {
   	/* Read word(s) into given location. 	*/
   for ( i = 0; i < ulNumBytes/sizeof(short); i++ )
		*( pwData + i ) = *( pwAddress + i );
  }
 else
  {
   	/* Write word(s) to specified location.	*/
   for ( i = 0; i < ulNumBytes/sizeof(short); i++ )
		*( pwAddress + i ) = *( pwData + i );
  }

 return( ERROR_SUCCESS );
}


/*******************************************************************************
**	   DwordIo():	Function to execute 32-bit memory I/O operations				**
**																										**
**		Arguments:  dwSessionID - number that describes I/O session					**
**						iTask - specifies operation to perform								**
**									  ( TRUE => Read, FALSE => Write )						**
**						ulOffset - physical address to manage								**
**						ulNumBytes - number of BYTES to manage								**
**						pulData - address to hold requested long(s)		 				**
**						iResGroup - especifies resource group for transfer				**
**									  ( TRUE => Secondary, FALSE => Primary )				**
**																										**
**		Return value:	ERROR_SUCCESS (See Winerror.h)									**
*******************************************************************************/
DWORD DwordIo(DWORD dwSessionID, BOOL iTask, ULONG ulOffset,
					ULONG ulNumBytes, PULONG pulData, BOOL iResGroup)
{
 ULONG	i;
 PULONG	pulAddress;

 if ( !iResGroup )
 	pulAddress = (PULONG)(Session[dwSessionID].PhysMem1);
 else
   pulAddress = (PULONG)(Session[dwSessionID].PhysMem2);

 if (pulAddress == NULL)              /* Check for invalid pointer before	*/
 	return ( ERROR_INVALID_HANDLE );  /* attempting any access.					*/

 pulAddress += ulOffset;     			 /* Start access at requested offset. 	*/

 if ( iTask )
  {
   	/* Read dword(s) into given location. 	*/
   for ( i = 0; i < ulNumBytes/sizeof(long); i++ )
		*( pulData + i ) = *( pulAddress + i );
  }
 else
  {
   	/* Write dword(s) to specified location.	*/
   for ( i = 0; i < ulNumBytes/sizeof(long); i++ )
		*( pulAddress + i ) = *( pulData + i );
  }

 return( ERROR_SUCCESS );
}


/*******************************************************************************
**	   SingleIo():	Function to execute single-precision memory I/O operations	**
**																										**
**		Arguments:  dwSessionID - number that describes I/O session					**
**						iTask - specifies operation to perform								**
**									  ( TRUE => Read, FALSE => Write )						**
**						ulOffset - physical address to manage								**
**						ulNumBytes - number of BYTES to manage								**
**						pfData - address to hold requested single(s)		 				**
**						iResGroup - especifies resource group for transfer				**
**									  ( TRUE => Secondary, FALSE => Primary )				**
**																										**
**		Return value:	ERROR_SUCCESS (See Winerror.h)									**
*******************************************************************************/
DWORD SingleIo(DWORD dwSessionID, BOOL iTask, ULONG ulOffset,
					ULONG ulNumBytes, PFLOAT pfData, BOOL iResGroup)
{
 ULONG	 i;
 PFLOAT	 pfAddress;

 if ( !iResGroup )
 	pfAddress = (PFLOAT)(Session[dwSessionID].PhysMem1);
 else
   pfAddress = (PFLOAT)(Session[dwSessionID].PhysMem2);

 if (pfAddress == NULL)              /* Check for invalid pointer before	*/
 	return ( ERROR_INVALID_HANDLE );  /* attempting any access.					*/

 pfAddress += ulOffset;     			 /* Start access at requested offset. 	*/

 if ( iTask )
  {
   	/* Read single(s) into given location. 	*/
   for ( i = 0; i < ulNumBytes/sizeof(float); i++ )
		*( pfData + i ) = *( pfAddress + i );
  }
 else
  {
   	/* Write single(s) to specified location.	*/
   for ( i = 0; i < ulNumBytes/sizeof(float); i++ )
		*( pfAddress + i ) = *( pfData + i );
  }

 return( ERROR_SUCCESS );
}


/*******************************************************************************
**	   DoubleIo():	Function to execute double-precision memory I/O operations	**
**																										**
**		Arguments:  dwSessionID - number that describes I/O session					**
**						iTask - specifies operation to perform								**
**									  ( TRUE => Read, FALSE => Write )						**
**						ulOffset - physical address to manage								**
**						ulNumBytes - number of BYTES to manage								**
**						pdData - address to hold requested double(s)		 				**
**						iResGroup - especifies resource group for transfer				**
**									  ( TRUE => Secondary, FALSE => Primary )				**
**																										**
**		Return value:	ERROR_SUCCESS (See Winerror.h)									**
*******************************************************************************/
DWORD DoubleIo(DWORD dwSessionID, BOOL iTask, ULONG ulOffset,
					ULONG ulNumBytes, PDOUBLE pdData, BOOL iResGroup)
{
 ULONG	 i;
 PDOUBLE	 pdAddress;

 if ( !iResGroup )
 	pdAddress = (PDOUBLE)(Session[dwSessionID].PhysMem1);
 else
   pdAddress = (PDOUBLE)(Session[dwSessionID].PhysMem2);

 if (pdAddress == NULL)              /* Check for invalid pointer before	*/
 	return ( ERROR_INVALID_HANDLE );  /* attempting any access.					*/

 pdAddress += ulOffset;     			 /* Start access at requested offset. 	*/

 if ( iTask )
  {
   	/* Read double(s) into given location. 	*/
   for ( i = 0; i < ulNumBytes/sizeof(double); i++ )
		*( pdData + i ) = *( pdAddress + i );
  }
 else
  {
   	/* Write double(s) to specified location.	*/
   for ( i = 0; i < ulNumBytes/sizeof(double); i++ )
		*( pdAddress + i ) = *( pdData + i );
  }

 return( ERROR_SUCCESS );
}


/*******************************************************************************
**	   PortIo():	Function to execute data transfers with 8-bit ports			**
**																										**
**		Arguments:  dwSessionID - number that describes I/O session					**
**						iTask - specifies operation to perform								**
**									  ( TRUE => Read, FALSE => Write )						**
**						ulAddress - physical address to manage								**
**						ulNumElements - number of data elements to transfer			**
**						pbData - address that holds session buffer		 				**
**																										**
**		Return value:	Success = 0																**
**							Failure = standard Windows error code 							**
*******************************************************************************/
DWORD PortIo(DWORD dwSessionID, BOOL iTask, ULONG ulAddress,
					ULONG ulNumElements, PBYTE pbData)
{
 ULONG	ulCount;
 register i;

 if ( iTask )
  {
   WINRT_CONTROL_ITEM	ReadCommand[] = {{INP_BA, 0, 0}};
   for (i=0; i < ulNumElements; i++)
    {
     ReadCommand[0].port = ulAddress;

		/* Process input command and copy bytes to specified location.		*/
	  if ( !WinRTProcessIoBuffer(Session[dwSessionID].Handle, ReadCommand,
   										sizeof(ReadCommand), &ulCount) )
		return ( GetLastError() );

	  *(pbData + i) = (BYTE)(ReadCommand[0].value);
    }
  }
 else
  {
   WINRT_CONTROL_ITEM	WriteCommand[] = {{OUTP_BA, 0, 0}};
   for (i=0; i < ulNumElements; i++)
    {
	  WriteCommand[0].port = ulAddress;
     WriteCommand[0].value = (ULONG)( *(pbData + i));

			/* Process output command and write byte to specified location.		*/
	  if ( !WinRTProcessIoBuffer(Session[dwSessionID].Handle, WriteCommand,
   										sizeof(WriteCommand), &ulCount) )
		return ( GetLastError() );
    }
  }

 return( ERROR_SUCCESS );
}


/*******************************************************************************
**	   PortIo_16():	Function to execute data transfers with 16-bit ports		**
**																										**
**		Arguments:  dwSessionID - number that describes I/O session					**
**						iTask - specifies operation to perform								**
**									  ( TRUE => Read, FALSE => Write )						**
**						ulAddress - physical address to manage								**
**						ulNumElements - number of data elements to transfer			**
**						pwData - address that holds session buffer		 				**
**																										**
**		Return value:	Success = 0																**
**							Failure = standard Windows error code 							**
*******************************************************************************/
DWORD PortIo_16(DWORD dwSessionID, BOOL iTask, ULONG ulAddress,
						ULONG ulNumElements, PUSHORT pwData)
{
 ULONG	ulCount;
 register i;

 if ( iTask )
  {
   WINRT_CONTROL_ITEM	ReadCommand[] = {{INP_WA, 0, 0}};
   for (i=0; i < ulNumElements; i++)
    {
     ReadCommand[0].port = ulAddress;

		/* Process input command and copy bytes to specified location.		*/
	  if ( !WinRTProcessIoBuffer(Session[dwSessionID].Handle, ReadCommand,
   										sizeof(ReadCommand), &ulCount) )
		return ( GetLastError() );

	  *(pwData + i) = (USHORT)(ReadCommand[0].value);
    }
  }
 else
  {
   WINRT_CONTROL_ITEM	WriteCommand[] = {{OUTP_WA, 0, 0}};
   for (i=0; i < ulNumElements; i++)
    {
	  WriteCommand[0].port = ulAddress;
     WriteCommand[0].value = (ULONG)( *(pwData + i));

			/* Process output command and write byte to specified location.		*/
	  if ( !WinRTProcessIoBuffer(Session[dwSessionID].Handle, WriteCommand,
   										sizeof(WriteCommand), &ulCount) )
		return ( GetLastError() );
    }
  }

 return( ERROR_SUCCESS );
}


/*******************************************************************************
**	   PortIo_32():	Function to execute data transfers with 32-bit ports		**
**																										**
**		Arguments:  dwSessionID - number that describes I/O session					**
**						iTask - specifies operation to perform								**
**									  ( TRUE => Read, FALSE => Write )						**
**						ulAddress - physical address to manage								**
**						ulNumElements - number of data elements to transfer			**
**						pulData - address that holds session buffer		 				**
**																										**
**		Return value:	Success = 0																**
**							Failure = standard Windows error code 							**
*******************************************************************************/
DWORD PortIo_32(DWORD dwSessionID, BOOL iTask, ULONG ulAddress,
						ULONG ulNumElements, PULONG pulData)
{
 ULONG	ulCount;
 register i;

 if ( iTask )
  {
   WINRT_CONTROL_ITEM	ReadCommand[] = {{INP_LA, 0, 0}};
   for (i=0; i < ulNumElements; i++)
    {
     ReadCommand[0].port = ulAddress;

		/* Process input command and copy bytes to specified location.		*/
	  if ( !WinRTProcessIoBuffer(Session[dwSessionID].Handle, ReadCommand,
   										sizeof(ReadCommand), &ulCount) )
		return ( GetLastError() );

	  *(pulData + i) = ReadCommand[0].value;
    }
  }
 else
  {
   WINRT_CONTROL_ITEM	WriteCommand[] = {{OUTP_LA, 0, 0}};
   for (i=0; i < ulNumElements; i++)
    {
	  WriteCommand[0].port = ulAddress;
     WriteCommand[0].value = *(pulData + i);

			/* Process output command and write byte to specified location.		*/
	  if ( !WinRTProcessIoBuffer(Session[dwSessionID].Handle, WriteCommand,
   										sizeof(WriteCommand), &ulCount) )
		return ( GetLastError() );
    }
  }

 return( ERROR_SUCCESS );
}


/*******************************************************************************
**		DriverControl(): Function to start and stop WinRT VxDs. (95 only)			**
**																										**
**		Arguments:  eTask - specifies operation to perform								**
**						dwDeviceNumber - device number to control							**
**																										**
**		Return value: dwErrorCode - standard Windows error code for		 			**
**												most recent error		 							**
*******************************************************************************/
DWORD DriverControl(enum DCtrlFunc eTask, DWORD dwDeviceNumber)
{
 HANDLE  	hVxDLdr;    					/* Handle to VxD Loader */
 CHAR    	szVxDName[MAX_PATH];
 CHAR    	szTemp[MAX_PATH];
 USHORT   	dwErrorCode;
 IOCTL_PKT 	DL_IOCTL_Pkt;
 DWORD		dwBytesReturned;
 HANDLE 		hVxdFile;               /* Handle to Find File session.	*/
 WIN32_FIND_DATA	FileData;

	/* Get file handle to VXDLDR VxD */
 hVxDLdr = CreateFile("\\\\.\\VXDLDR", 0,0,0,0,0,0);
 if ( hVxDLdr == INVALID_HANDLE_VALUE )
    return( GetLastError() );

 switch ( eTask )
  {
   case (Start_Driver):

	  	/* Build path & name of VxD.  First, assume that VxD is in the system 	*/
	  	/* folder (e.g. C:\WINDOWS\SYSTEM).  If this fails, look for the VxD in	*/
      /* the VMM32 directory (e.g. C:\WINDOWS\SYSTEM\VMM32).						*/

 	  GetSystemDirectory(szVxDName, MAX_PATH);
 	  sprintf(szTemp, PLUG_N_PLAY_DEVICE, dwDeviceNumber);
     lstrcat(szVxDName, szTemp);

		/* Verify that the VxD exists.  If it does not exist, exit function.		*/
     hVxdFile = FindFirstFile( (LPCTSTR)(szVxDName), &FileData);
     if ( hVxdFile == INVALID_HANDLE_VALUE )
     	{
 	  	 GetSystemDirectory(szVxDName, MAX_PATH);
 	  	 sprintf(szTemp, ISA_DEVICE, dwDeviceNumber);
     	 lstrcat(szVxDName, szTemp);

       hVxdFile = FindFirstFile( (LPCTSTR)(szVxDName), &FileData);
     	 	if ( hVxdFile == INVALID_HANDLE_VALUE )
				return( GetLastError() );
      }
     FindClose(hVxdFile);

     	/* Call the Vxd Loader to load our Vxd. */
     DeviceIoControl(hVxDLdr, VXDLDR_APIFUNC_LOADDEVICE, (LPVOID)szVxDName,
						lstrlen(szVxDName)+1, (LPVOID)&DL_IOCTL_Pkt,
                  sizeof(IOCTL_PKT), &dwBytesReturned, NULL);

		/* Save function call outcome if non-trivial error is detected.	*/
     	if (DL_IOCTL_Pkt.Res_DL_Pkt.W32IO_ErrorCode != VXDLDR_ERR_DUPLICATE_DEVICE)
     		dwErrorCode = DL_IOCTL_Pkt.Res_DL_Pkt.W32IO_ErrorCode;
      else
      	dwErrorCode = ERROR_SUCCESS;
     break;

   case (Stop_Driver):
      /* Create the Module name (this is case sensitive). */
     sprintf((char *)(DL_IOCTL_Pkt.Res_DL_Pkt.W32IO_ModuleName), "WRTDEV%d", dwDeviceNumber);
     DL_IOCTL_Pkt.Res_DL_Pkt.W32IO_DeviceID = (USHORT)-1;

      /* Call VxD Loader to Unload our Vxd. */
     DeviceIoControl(hVxDLdr, VXDLDR_APIFUNC_UNLOADDEVICE, (LPVOID)&DL_IOCTL_Pkt,
     						sizeof(IOCTL_PKT), (LPVOID)&dwErrorCode, sizeof(USHORT),
            			&dwBytesReturned, NULL);

     /* Verify error code is not "Device already unloaded".	*/
     if (dwErrorCode == VXDLDR_ERR_NO_SUCH_DEVICE)
     	dwErrorCode = ERROR_SUCCESS;
   break;
  } /* switch */

	/* Done with the Loader. */
 CloseHandle(hVxDLdr);

 return( (DWORD)(dwErrorCode) );
}


/*******************************************************************************
**   	ConfigureDevice() - This function serves as an interface between			**
**									LabVIEW and the DeviceToRegistry function.         **
**																										**
**		Arguments:  dwDeviceNum - Number of device to configure						**
**						szDeviceID - name of device to configure							**
**						Config - LabVIEW-supplied device data description	         **
**						AtDefault - LabVIEW-supplied default device description.    **
**									  If a configuration item is at its default value	**
**									  that element of the AtDefault array is TRUE, 		**
**									  otherwise, the element is FALSE.						**
**																										**
**		Return value:	Success = 0																**
**							Failure = standard Windows error code 							**
**										 (defined in winerror.h)								**
*******************************************************************************/
DWORD ConfigureDevice(uInt32 dwDeviceNum, CStr szDeviceID, uInt32 *Config,
								uInt8 *AtDefault)
{
 DEVICE device;							/*	device data description	*/

     /* Initialize all items to not present.                            **
     ** Items which are not present will not be written to              **
     ** the registry, this allows the drivers to apply                  **
     ** appropriate default values for the systems they are running on.	*/

 device.PortAdrs.bItemEmpty      = TRUE;         /* Section 0 Data	*/
 device.PortCnt.bItemEmpty       = TRUE;
 device.MemAdrs.bItemEmpty       = TRUE;
 device.MemCnt.bItemEmpty        = TRUE;
 device.IntNum.bItemEmpty        = TRUE;
 device.IntLevel.bItemEmpty      = TRUE;
 device.IntShare.bItemEmpty      = TRUE;
 device.IntMode.bItemEmpty       = TRUE;
 device.BusNum.bItemEmpty        = TRUE;
 device.BusType.bItemEmpty       = TRUE;
 device.IoIncrem.bItemEmpty      = TRUE;
 device.AllowConflict.bItemEmpty = TRUE;
 device.Disable.bItemEmpty       = TRUE;
 device.PortAdrs1.bItemEmpty     = TRUE;         /* Section 1 Data	*/
 device.PortCnt1.bItemEmpty      = TRUE;
 device.MemAdrs1.bItemEmpty      = TRUE;
 device.MemCnt1.bItemEmpty       = TRUE;

 if ( !(*AtDefault) )
  {
   device.PortAdrs.bItemEmpty = FALSE;        /*  If device item is not at	*/
   device.PortAdrs.dwItemValue = *(Config);   /*  its default value, initialize	*/
  }                                           /*  it properly.					*/

 if ( !(*(AtDefault + 1)) )
  {
   device.PortCnt.bItemEmpty = FALSE;
   device.PortCnt.dwItemValue = *(Config + 1);
  }

 if ( !(*(AtDefault + 2)) )
  {
   device.MemAdrs.bItemEmpty = FALSE;
   device.MemAdrs.dwItemValue = *(Config + 2);
  }

 if ( !(*(AtDefault + 3)) )
  {
   device.MemCnt.bItemEmpty = FALSE;
   device.MemCnt.dwItemValue = *(Config + 3);
  }

 if ( !(*(AtDefault + 4)) )
  {
   device.IntNum.bItemEmpty = FALSE;
   device.IntNum.dwItemValue = *(Config + 4);
  }

 if ( !(*(AtDefault + 5)) )
  {
   device.IntLevel.bItemEmpty = FALSE;
   device.IntLevel.dwItemValue = *(Config + 5);
  }

 if ( !(*(AtDefault + 6)) )
  {
   device.IntShare.bItemEmpty = FALSE;
   device.IntShare.dwItemValue = *(Config + 6);
  }

 if ( !(*(AtDefault + 7)) )
  {
   device.IntMode.bItemEmpty = FALSE;
   device.IntMode.dwItemValue = *(Config + 7);
  }

 if ( !(*(AtDefault + 8)) )
  {
   device.BusNum.bItemEmpty = FALSE;
   device.BusNum.dwItemValue = *(Config + 8);
  }

 if ( !(*(AtDefault + 9)) )
  {
   device.BusType.bItemEmpty = FALSE;
   device.BusType.dwItemValue = *(Config + 9);
  }

 if ( !(*(AtDefault + 10)) )
  {
   device.IoIncrem.bItemEmpty = FALSE;
   device.IoIncrem.dwItemValue = *(Config + 10);
  }

 if ( !(*(AtDefault + 11)) )
  {
   device.AllowConflict.bItemEmpty = FALSE;
   device.AllowConflict.dwItemValue = *(Config + 11);
  }

 if ( !(*(AtDefault + 12)) )
  {
   device.Disable.bItemEmpty = FALSE;
   device.Disable.dwItemValue = *(Config + 12);
  }

 if ( !(*(AtDefault + 13)) )
  {
   device.PortAdrs1.bItemEmpty = FALSE;
   device.PortAdrs1.dwItemValue = *(Config + 13);
  }

 if ( !(*(AtDefault + 14)) )
  {
   device.PortCnt1.bItemEmpty = FALSE;
   device.PortCnt1.dwItemValue = *(Config + 14);
  }

 if ( !(*(AtDefault + 15)) )
  {
   device.MemAdrs1.bItemEmpty = FALSE;
   device.MemAdrs1.dwItemValue = *(Config + 15);
  }

 if ( !(*(AtDefault + 16)) )
  {
   device.MemCnt1.bItemEmpty = FALSE;
   device.MemCnt1.dwItemValue = *(Config + 16);
  }

	/* Build VxD name string.	*/
 sprintf(device.szVxdName, "WRTDEV%d", dwDeviceNum);    /* Win95 VxD name.		*/
 lstrcpy(device.szDeviceName, (LPCSTR)(szDeviceID) );   /* WinNT device name. */
 device.dwDeviceNumber = dwDeviceNum;

 if (DriverToRegistry(&device))
 	 return (GetLastError());
 else
    return (DeviceToRegistry(&device));
}


/*******************************************************************************
**   	DeviceToRegistry() - Writes device data to Registry. 			            **
**																										**
**		Arguments:  pDevice - device data description                           **
**																										**
**		Return value:	Success = 0																**
**							Failure = standard Windows error code 							**
**										 (defined in winerror.h)								**
*******************************************************************************/
DWORD DeviceToRegistry(PDEVICE pDevice)
{
 HKEY hkReg;                         	  /* For Registry key	*/
 CHAR szReg[MAX_REGISTRY_NAME];       	  /* Registry name buffer	*/
 DWORD dwDisp;                       	  /* For Registry disposition value	*/
 DWORD dwErrorCode;


 if (lstrlen(pDevice->szDeviceName) == 0)
    return( ERROR_FILE_NOT_FOUND );

     /* Create Registry key for device Section0	*/
 lstrcpy(szReg, DRIVER_REGISTRY_BASE);
 lstrcat(szReg, "\\");
 lstrcat(szReg, pDevice->szDeviceName);
 lstrcat(szReg, "\\Parameters\\Section0");

 dwErrorCode = RegCreateKeyEx(HKEY_LOCAL_MACHINE, szReg, 0, NULL,
                REG_OPTION_NON_VOLATILE,KEY_ALL_ACCESS, NULL, &hkReg, &dwDisp);
 if ( dwErrorCode != ERROR_SUCCESS )
    return( dwErrorCode );         /* Catch failure creating Registry key.	*/

  /* Place each device parameter in the Registry (Section 0 params only).	*/
 if ( UpdateRegValue(hkReg, "IoPortAddress",  &(pDevice->PortAdrs))     ||
  	   UpdateRegValue(hkReg, "IoPortCount",    &(pDevice->PortCnt))      ||
      UpdateRegValue(hkReg, "MemoryAddress",  &(pDevice->MemAdrs))      ||
      UpdateRegValue(hkReg, "MemoryCount",    &(pDevice->MemCnt))       ||
      UpdateRegValue(hkReg, "Interrupt",      &(pDevice->IntNum))       ||
      UpdateRegValue(hkReg, "InterruptLevel", &(pDevice->IntLevel))     ||
      UpdateRegValue(hkReg, "InterruptMode",  &(pDevice->IntMode))      ||
      UpdateRegValue(hkReg, "InterruptShare", &(pDevice->IntShare))     ||
      UpdateRegValue(hkReg, "BusNumber",      &(pDevice->BusNum))       ||
      UpdateRegValue(hkReg, "BusType",        &(pDevice->BusType))      ||
      UpdateRegValue(hkReg, "IoIncrement",    &(pDevice->IoIncrem))     ||
      UpdateRegValue(hkReg, "AllowConflicts", &(pDevice->AllowConflict))||
      UpdateRegValue(hkReg, "Disable",        &(pDevice->Disable)) )
    {    													/* Error on Registry Set.	*/
     RegCloseKey(hkReg);            			   /* Close Registry Section 0	*/
     return( GetLastError() );                  /* and return error code.	*/
    }

 RegCloseKey(hkReg);                 /* Close Registry Section 0	*/

 	/* If any parameters for a second section of I/O for this device, these  **
   ** are saved in "Section1" key under "Parameters".                       **
   ** Create Registry key for new device Section1 - even if no values.	    */

 lstrcpy(szReg, DRIVER_REGISTRY_BASE);
 lstrcat(szReg, "\\");
 lstrcat(szReg, pDevice->szDeviceName);
 lstrcat(szReg, "\\Parameters\\Section1");

 dwErrorCode = RegCreateKeyEx(HKEY_LOCAL_MACHINE, szReg, 0, NULL,
               REG_OPTION_NON_VOLATILE,KEY_ALL_ACCESS, NULL, &hkReg, &dwDisp);
 if ( dwErrorCode != ERROR_SUCCESS )
  	 return( dwErrorCode );         /* Catch failure creating Registry key.	*/

  	/* Enter any parameters for Section1	*/
 if ( UpdateRegValue(hkReg, "IoPortAddress",  &(pDevice->PortAdrs1)) ||
   	UpdateRegValue(hkReg, "IoPortCount",    &(pDevice->PortCnt1))  ||
     	UpdateRegValue(hkReg, "MemoryAddress",  &(pDevice->MemAdrs1))  ||
      UpdateRegValue(hkReg, "MemoryCount",    &(pDevice->MemCnt1)) )
 	 {     											  	/* Error on Registry Set  */
     RegCloseKey(hkReg);             	  		/* Close Reg Section 1,   */
     return( GetLastError() );           		/* get error and return	  */
 	 }

 RegCloseKey(hkReg);                 	  /* Close Reg Section 1	 */
 return( ERROR_SUCCESS );                /* Everything OK	*/
}


/*******************************************************************************
**   UpdateRegValue() - Updates individual Registry entry.							**
**   Arguments:  hKey     - open registry key										   **
**            	  pszName  - name of key value to set									**
**            	  pItem    - Item structure to set                             **
**                                                                            **
**		Return value:	Success = FALSE														**
**							Failure = TRUE															**
*******************************************************************************/
DWORD UpdateRegValue(HKEY hKey, PCHAR pszName, PDEVICE_ITEM pItem)
{
 DWORD dwErrorCode;
 DWORD dwValue, dwLength;

	/* If item is empty, verify current Registry value. */
 if ( pItem->bItemEmpty )
   {
	 dwLength = sizeof(dwValue);
    if ( RegQueryValueEx(hKey, pszName, 0, NULL, (LPBYTE)&dwValue, &dwLength) )
 	 	return( FALSE );          /* Value is at default, do not overwrite. */
    else
	   dwErrorCode = RegDeleteValue(hKey, pszName);
   }

 else      /* If item is not empty, write new value to Registry.	*/
   {
 	 dwErrorCode = RegSetValueEx(hKey, pszName, 0, REG_DWORD,
	  				(BYTE *CONST)(&pItem->dwItemValue), sizeof(DWORD));
   }

 return( dwErrorCode != ERROR_SUCCESS );
}


/*******************************************************************************
**   DeleteDevice() - Deletes ALL information corresponding to the         	**
**								  specified device from the Registry.						**
**																										**
**		Arguments:	szDeviceID - identifies I/O device to control					**
**                                                                            **
**		Return value:	Success = 0																**
**							Failure = standard Windows error code 							**
**										 (defined in winerror.h)								**
*******************************************************************************/
DWORD DeleteDevice(CStr szDeviceID)
{
 HKEY hkReg;                         	  		/* For Registry key	*/
 DWORD dwErrorCode;

	/* Open handle to Device Key.	*/
 dwErrorCode = RegOpenKeyEx(HKEY_LOCAL_MACHINE, DRIVER_REGISTRY_BASE,
 									 0,KEY_ALL_ACCESS, &hkReg);
 if ( dwErrorCode != ERROR_SUCCESS )
    return( dwErrorCode );         /* Catch failure opening Registry key.	*/

	/* Delete Device Key.	*/
 dwErrorCode = RegDeleteKey( hkReg,(LPCSTR)(szDeviceID) );
 RegCloseKey(hkReg);                 /* Finished accessing Registry.	*/

 return( dwErrorCode );
}


/*******************************************************************************
**		GetSessionConfiguration():	This function returns the current Registry	**
**									settings for the specified session.						**
**																										**
**		Arguments:	dwSessionID - number that describes I/O session					**
**						Config - returns array with session configuration values	   **
**																										**
**		Return value:	Success = 0																**
**							Failure = standard Windows error code 							**
**										 (defined in winerror.h)								**
*******************************************************************************/
GetSessionConfiguration(DWORD dwSessionID, uInt32 *Config)
{
 WINRT_CONFIGURATION	Settings;
 DWORD dwLength;

 if ( !WinRTGetConfiguration(Session[dwSessionID].Handle, &Settings, &dwLength) )
 	return ( GetLastError() );

 *(Config) = Settings.portMin;                   		/* Section 0 Data	*/

 if ( Settings.portMax )// != Settings.portMin )
 	*(Config + 1) = (Settings.portMax - Settings.portMin) + 1;
// else
// 	*(Config + 1) = 0;

 *(Config + 2) = Settings.memMin;

 if ( Settings.memMax )//!= Settings.memMin )
 	*(Config + 3) = (Settings.memMax - Settings.memMin) + 1;
// else
// 	*(Config + 3) = 0;

 if ( Settings.interruptVector > 50 )          	/* Catch "undefined" numbers.	*/
   *(Config + 4) = 0;                           /* It generally indicates no	*/
 else                                           /* interrupts were assigned 	*/
 	*(Config + 4) = Settings.interruptVector;    /* to this device.				*/

 if ( Settings.interruptLevel > 50 )
   *(Config + 5) = 0;
 else
 	*(Config + 5) = Settings.interruptLevel;

 *(Config + 8) = Settings.busNumber;
 *(Config + 9) = Settings.busType;
 *(Config + 10) = Settings.ioIncrement;

 *(Config + 13) = Settings.portMin1;                   		/* Section 1 Data	*/

 if ( Settings.portMax1 )//!= Settings.portMin1 )
 	*(Config + 14) = (Settings.portMax1 - Settings.portMin1) + 1;
// else
// 	*(Config + 14) = 0;

 *(Config + 15) = Settings.memMin1;

 if ( Settings.memMax1 )//!= Settings.memMin1 )
   *(Config + 16) = (Settings.memMax1 - Settings.memMin1) + 1;
// else
// 	*(Config + 16) = 0;

 return ( ERROR_SUCCESS );
}


/*******************************************************************************
**	Search():	Used by AllocateNewSession to find available device.				**
**     			Searches for Key in List.													**
**	Arguments:	Key - number to search for													**
**					List - array to be searched for value									**
**					nElements - number of elements in List									**
**																										**
**	Returns:	Success - index of Key in List												**
**				Failure - -1                                                      **
*******************************************************************************/
int Search(int iKey, int List[], int nElements)
{
 register i = 0;

 while ( (List[i] != iKey) && ( i < nElements) ) i++;

 if (List[i] != iKey)
 	return(-1);

 return(i);
}


/*******************************************************************************
**	Delay():	This function waits for the specified number of milliseconds		**
**     		and then returns to its caller.												**
**	Arguments:	dwMilliseconds - number of milliseconds to wait						**
**																										**
**	Returns:	none																					**
*******************************************************************************/
void Delay(DWORD dwMilliseconds)
{
 DWORD start;

 start = GetTickCount();
 while ( (GetTickCount() - start) < dwMilliseconds );

 return;
}


/*******************************************************************************
**	   MemControl():	Function to open and close memory maps for any session.	**
**																										**
**		Arguments:  dwSessionID - number that describes I/O session					**
**						eTask - type of operation to perform								**
**						BaseMem1 - base address of primary memory block				   **
**						CountMem1 - length of primary memory block			   		**
**						BaseMem2 - base address of secondary memory block				**
**						CountMem2 - length of secondary memory block			   		**
**																										**
**		Return value:	Success = 0																**
**							Failure = standard Windows error code 							**
**										 (defined in winerror.h)								**
*******************************************************************************/
DWORD MemControl(DWORD dwSessionID, enum MemCommand eTask, DWORD ulBaseMem1,
						DWORD ulCountMem1, DWORD ulBaseMem2, DWORD ulCountMem2)
{
 ULONG	 		   ulCount;
 WINRT_MEMORY_MAP Region;        		/* Description of memory block.  */

 switch( eTask )
  {
   case ( MapMemory ):
    	{  /* If session has memory resources, map them.	*/
    	 if ( ulCountMem1 > 0 )
     		{
 			 Region.address = (PVOID)(ulBaseMem1);
 			 Region.length  = ulCountMem1;

   			/* Map in primary memory range.  */
          if ( !WinRTMapMemory(Session[dwSessionID].Handle, &Region,
      								&Session[dwSessionID].PhysMem1, &ulCount) )
          	return( GetLastError() );
     		}

    	 if ( ulCountMem2 > 0 )
     		{
 			 Region.address = (PVOID)(ulBaseMem2);
 			 Region.length  = ulCountMem2;

   			/* Map in secondary memory range.  */
 			 if ( !WinRTMapMemory(Session[dwSessionID].Handle, &Region,
      								&Session[dwSessionID].PhysMem2, &ulCount) )
          	return( GetLastError() );
     		}
    	 break;
      }

   case ( UnmapMemory ):
   	{	/* Unmap memory allocated to the session.	*/
    	 if ( ulCountMem1 > 0 )
     		{	/* Unmap primary memory range.  */
	 		 if ( !WinRTUnMapMemory(Session[dwSessionID].Handle,
      								Session[dwSessionID].PhysMem1, &ulCount) )
          	return( GetLastError() );

          	/*	Invalidate memory pointer.	*/
      	 Session[dwSessionID].PhysMem1 = NULL;
     		}

    	 if ( ulCountMem2 > 0 )
     		{	/* Unmap secondary memory range.  */
	 		 if ( !WinRTUnMapMemory(Session[dwSessionID].Handle,
      								Session[dwSessionID].PhysMem2, &ulCount) )
          	return( GetLastError() );

            /*	Invalidate memory pointer.	*/
      	 Session[dwSessionID].PhysMem2 = NULL;
         }
    	 break;
      }

   default: break;
  }

 return( ERROR_SUCCESS );
}


/*******************************************************************************
**	   WaitForInterrupt():	Function to asynchronuosly wait for an interrupt.	**
**																										**
**		Arguments:  dwSessionID - number that describes I/O session					**
**						iOccurrence - LabVIEW occurrence to be set						**
**																										**
**		Return value:	Success = 0																**
**							Failure = standard Windows error code 							**
**										 (defined in winerror.h)								**
*******************************************************************************/
DWORD WaitForInterrupt(DWORD dwSessionID, int32 iOccurrence)
{
 HANDLE hSessionDevice;						  /* Handle to session's WinRT device.	*/
 HANDLE hIsrThread;						  	  /* Handle to new IsrFunction thread.	*/
 DWORD  dwThreadId;        				  /* ID of IsrFunction thread.			*/
 char szDeviceName[MAX_WINRT_NAME];

 	/* Check for IsrThread initialization.	*/
 if (Session[dwSessionID].IsrEvent == NULL )
 	{
		/* Make sure that new thread will not be terminated.	*/
    Session[dwSessionID].TerminateIsr = FALSE;

		/* Create thread synchronization object.  The event object is set up		*/
      /* as an auto-reset event initialized to a nonsignaled state.				*/
    Session[dwSessionID].IsrEvent = CreateEvent(NULL, FALSE, FALSE, NULL);
    if (Session[dwSessionID].IsrEvent == NULL)
    	 return( GetLastError() );

	 	/*	Generate DeviceName string with session device number. */
      /* Win NT device is WinRTdev*; Win95 device is WRTdev*.	 */
    sprintf(szDeviceName, "\\\\.\\WRTdev%d", Session[dwSessionID].Device);

	 	/*	Obtain a new handle to session's device (for the new thread).	*/
	 hSessionDevice = CreateFile( (LPCTSTR)(szDeviceName), 0,
 						FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 0, NULL);

		/* Catch invalid handle.	*/
	 if ( hSessionDevice == INVALID_HANDLE_VALUE )
    	{
       CloseHandle(Session[dwSessionID].IsrEvent);
       Session[dwSessionID].IsrEvent = NULL;
	  	 return( GetLastError() );
      }

	   /* Save interrupt settings in Session Table structure.	*/
	 Session[dwSessionID].Occurrence = iOccurrence;
    ID = dwSessionID;

	   /* Create IsrFunction thread in suspended state.	*/
	 hIsrThread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)IsrFunction,
										hSessionDevice, CREATE_SUSPENDED, &dwThreadId);

	   /* Check thread creation.	*/
	 if ( hIsrThread == NULL )
    	{
       CloseHandle(hSessionDevice);            	  /* Failed to create thread.	*/
       CloseHandle(Session[dwSessionID].IsrEvent);
       Session[dwSessionID].IsrEvent = NULL;
       return ( GetLastError() );
      }

	   /* Set the priority for the new thread.	*/
	 SetThreadPriority(hIsrThread, THREAD_PRIORITY_TIME_CRITICAL);

	    /* Get a handle to the LabVIEW window to be notified of interrupt.	*/
	 Session[dwSessionID].WndHandle = FindWindow(NULL,(LPCTSTR)("LabVIEW"));
	 if ( Session[dwSessionID].WndHandle == NULL )
    	{
       Session[ID].TerminateIsr = TRUE;
       ResumeThread(hIsrThread);
       CloseHandle(hIsrThread);
	 	 return( GetLastError() );
      }

	   /* Start up IsrFunction thread and close setup handle.	*/
    ResumeThread(hIsrThread);
    CloseHandle(hIsrThread);   //It may be wrong to close hIsrThread at this time
   }		//because there will be no handles associated with the thread object and
         //when that happens Windows can destroy the thread object.
 else
 	{	/* If thread has been initialized, wake it up to service interrupts.	*/
    ID = dwSessionID;
    SetEvent(Session[dwSessionID].IsrEvent);
   }

 return(GetLastError());
}


/*******************************************************************************
**	   IsrFunction():	Thread to service session interrupts.							**
**																										**
**		Arguments:  hSessionDevice - handle to session's WinRT device				**
**																										**
**		Return value:	Success = 0																**
**							Failure = standard Windows error code 							**
**										 (defined in winerror.h)								**
*******************************************************************************/
DWORD IsrFunction(HANDLE hSessionDevice)
{
 DWORD dwLength;
 WINRT_CONTROL_ITEM CommandBuffer[2];

   /*	Set up command buffer with universal interrupt settings.	 */
	/* Driver will ALWAYS indentify the interrupt as ours and	 */
   /* will NOT CLEAR the interrupt upon its occurrence.			 */
 CommandBuffer[0].command = INTRP_ID_ALWAYS;
 CommandBuffer[1].command = INTRP_CLEAR_NOP;

Loop:
 if (Session[ID].TerminateIsr)
  {
   CloseHandle(hSessionDevice);
   CloseHandle(Session[ID].IsrEvent);
	CloseHandle(Session[ID].WndHandle);
   Session[ID].IsrEvent = NULL;
   Session[ID].WndHandle = NULL;
 	ExitThread(ERROR_SUCCESS);
  }
 else
  {	 /* Make thread wait for interrupt occurrence.	*/
	WinRTWaitForInterrupt(hSessionDevice, CommandBuffer, sizeof(CommandBuffer), &dwLength);

	    /* Notify LabVIEW of interrupt.	*/
	SendMessage(Session[ID].WndHandle, WM_USER+2, 0, Session[ID].Occurrence);

       /* This if is a debugging aid.  It has no other purpose.	*/
   if (Session[ID].IsrEvent == NULL)
   	Beep(500, 500);

    	 /* Suspend thread execution.	*/
	if (WaitForSingleObject(Session[ID].IsrEvent, INFINITE) == WAIT_FAILED)
   	Session[ID].TerminateIsr = TRUE;
   goto Loop;
  }

 return (ERROR_SUCCESS);
}


/*******************************************************************************
**   DriverToRegistry() - Writes (VxD) driver data to registry.               **
**   Inputs:  pDevice - device data description                               **
**																										**
**   Return: FALSE - if successful                                            **
**				 TRUE - if function failed								                  **
*******************************************************************************/
BOOLEAN DriverToRegistry(PDEVICE pDevice)
{
 HKEY  hkReg;                        // for Registry key
 DWORD dwDisp;                       // for Reg disposition value
 LONG  Error;                        // for error code
 HKEY  hkVxdReg;                     // for Registry key
 DWORD dwStart;                      // start type
 DRIVER driver;                      // driver data

 
	/* Set up the driver information.  This determines when and how the VxDs	**
   ** will be started.  For us, VxDs will always be started manually.			*/
 driver.dwWinRTStart = SERVICE_DEMAND_START;

	/* Open (or create if it does not exist) the WinRT Reg.	*/
 Error = RegCreateKeyEx(HKEY_LOCAL_MACHINE, DRIVER_REGISTRY_BASE, 0, NULL,
                           REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS,
                           NULL, &hkReg, &dwDisp);
 if (Error != ERROR_SUCCESS)
   return(TRUE);

	/* Windows 95 requires a set up for each device (implemented as an individual VxD)  **
   ** Save the start code in both registry branches.												*/
 if ((Error = RegSetValueEx(hkReg, "Start", 0, REG_DWORD,
        (LPBYTE)(driver.dwWinRTStart), sizeof(DWORD))) != ERROR_SUCCESS)
 	{
    	/* Registry Set value error.	*/
    RegCloseKey(hkReg);
    return(TRUE);
   }

 RegCloseKey(hkReg);		 /* Finished accessing Reg.	*/

	/* For each device add key to VxD start branch.	*/
	/* Open the Vxd branch.	*/
 if ((Error = RegCreateKeyEx(HKEY_LOCAL_MACHINE, VXD_REGISTRY_BASE, 0, NULL,
                               REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS,
                               NULL, &hkReg, &dwDisp)) != ERROR_SUCCESS)
	return(TRUE);

	/* Set start based on Win95 Vxd required values. Although we are going to  **
   ** configure all of our devices to start manually, we keep the code to		**
   ** handle automatic startup for possible future use.								*/
 if ((driver.dwWinRTStart == SERVICE_BOOT_START) ||
       (driver.dwWinRTStart == SERVICE_SYSTEM_START))
    dwStart = 0;    /* Start at boot. */
 else
    dwStart = 4;    /* Do not start at boot. */

 if (dwStart == 0)
  {
   CHAR   szTempName[MAX_DEVICE_NAME+1];

  	/* Create Key for this device. */
   if (RegCreateKeyEx(hkReg, (LPCSTR)pDevice->szVxdName, 0, NULL,
   							REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS,
                        NULL, &hkVxdReg, &dwDisp) != ERROR_SUCCESS)
   	{
   	 /* Registry Create error. */
   	 RegCloseKey(hkVxdReg);
   	 RegCloseKey(hkReg);
   	 return(TRUE);
   	}

   /* Set start value. */
   if (RegSetValueEx(hkVxdReg, "Start", 0, REG_DWORD,
                        (LPBYTE)(&dwStart), sizeof(DWORD)) != ERROR_SUCCESS)
  		{
       /* Registry set value error. */
       RegCloseKey(hkVxdReg);
       RegCloseKey(hkReg);
       return(TRUE);
      }

   /* Set Vxd device name - prepend an asterisk to force Win95 to search the	**
   ** VMM32 folder for the VxD.												 				*/
   strcpy(szTempName, "*");
   strcat(szTempName, pDevice->szVxdName);
   if (RegSetValueEx(hkVxdReg, "StaticVxD", 0, REG_SZ, (LPBYTE)(LPCSTR)(szTempName),
                        lstrlen(pDevice->szVxdName)) != ERROR_SUCCESS)
   	{
       /* Registry Set value error. */
       RegCloseKey(hkVxdReg);
       RegCloseKey(hkReg);
       return(TRUE);
      }

   RegCloseKey(hkVxdReg);  // Finished accessing  this VxD Reg
  }
 else
  {
   /* Start code does work correctly on some Windows 95 builds,					**
   ** setting to 4 will not stop loading, so just delete the entry.				*/
   RegDeleteKey(hkReg, (LPCSTR)pDevice->szVxdName);
  }

 RegCloseKey(hkReg); // Finished accessing Reg

 return(FALSE);
}

